/*-
 * Copyright (c) 2018 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

// Package log implements a client for the Varnish native logging
// interface.
package log

/*
#cgo pkg-config: varnishapi
#include "log.h"
*/
import "C"

// import "uplex.de/varnishapi/internal/pkg/vsm"

import (
	"errors"
	"fmt"
	"os"
	"regexp"
	"unsafe"
)

// Possible values for type TxType, which classifies a log
// transaction.
const (
	// Unknown transaction type.
	TxUnknown = TxType(C.VSL_t_unknown)

	// The log of a session, which stands for the "conversation"
	// that Varnish has with a single client connection,
	// comprising one or more requests/responses.
	Sess = TxType(C.VSL_t_sess)

	// A client request/response log.
	Req = TxType(C.VSL_t_req)

	// A backend request/response log.
	BeReq = TxType(C.VSL_t_bereq)

	// A raw transaction with only one Record, such as the log of
	// a backend health check.
	TxRaw = TxType(C.VSL_t_raw)
)

// TxType is a classifier for a log transaction, indicating if it is
// the log of a client or backend request/response, a Varnish session,
// or a raw transaction. Possible values for TxType are shown in the
// constants above.
type TxType uint8

// Possible values for a Reason, which classifies the cause of the
// event that was logged.
const (
	// Unknown reason.
	ReasonUnknown = Reason(C.VSL_r_unknown)

	// The start of an HTTP/1 session.
	HTTP1 = Reason(C.VSL_r_http_1)

	// A client request was received.
	Rxreq = Reason(C.VSL_r_rxreq)

	// An ESI subrequest.
	ESI = Reason(C.VSL_r_esi)

	// A request restart.
	Restart = Reason(C.VSL_r_restart)

	// A backend request resulting from a pass (cache lookup was
	// bypassed).
	Pass = Reason(C.VSL_r_pass)

	// A synchronous backend fetch.
	Fetch = Reason(C.VSL_r_fetch)

	// A backend fetch in the background.
	Bgfetch = Reason(C.VSL_r_bgfetch)

	// A pipe (exchanging bytes with a backend until one or the
	// other side closes the connection).
	Pipe = Reason(C.VSL_r_pipe)
)

// Reason is a classifier for the cause of the event that was logged.
// Possible values for Reason are shown in the constants above.
type Reason uint8

// Possible values for a Grouping, which indicates how transactions
// are aggregated when read from the log. The default Grouping is
// VXID.
const (
	// Transactions are not grouped, nor are records grouped into
	// transactions -- every transaction has exactly one
	// record. Records unrelated to requests, responses and
	// sessions, such as backend health checks, are only reported
	// in this mode.
	GRaw = Grouping(C.VSL_g_raw)

	// In the default mode, transactions are not grouped -- each
	// read returns one transaction for a request/response (client
	// or backend) or session. Non-transactional data, such as
	// backend health checks, are not reported in this mode. VXID
	// is the unique ID generated by Varnish for each transaction.
	VXID = Grouping(C.VSL_g_vxid)

	// Group by client request. A group contains the client
	// request and any other transactions that it initiated, such
	// as backend requests, ESI subrequests and restarts,
	Request = Grouping(C.VSL_g_request)

	// Group all transactions initiated by a client
	// connection. These groups can be much larger than the
	// others; they may occupy significantly more memory, and log
	// reads with this mode may take much longer.
	Session = Grouping(C.VSL_g_session)
)

// Grouping mode determines how transactions are aggregated when read
// from the log. Depending on the mode, transactions in a group may
// form a hierarchy. Possible values for Grouping are shown in the
// constants above. The default Grouping is VXID.
type Grouping uint8

// Possible values for RecordType, which classifies the data in a
// Record.
const (
	// The Record is part of the log for a client
	// request/response.
	Client = RecordType('c')

	// Part of the log for a backend request/response.
	Backend = RecordType('b')

	// Not part of the log of a backend or client
	// transaction. Examples are records with tags such as
	// "Backend_health", "CLI" and so forth.
	None = RecordType('-')
)

// RecordType is a classifier for the data in a Record, indicating if
// it is part of the log for a client request/response, a backend
// request/repsonse, or neither.  Corresponds to the indicator 'b',
// 'c' or '-' in the fourth column of verbose varnishlog
// output. Possible values for RecordType are shown in the constants
// above.
type RecordType uint8

func tags() []string {
	var tags []string
	if uint8(C.slt_max()) > ^uint8(0) {
		panic("SLT__MAX > max uint8")
	}
	max := int(C.slt_max())
	for i := 0; i < max; i++ {
		if C.VSL_tags[i] == nil {
			tags = append(tags, "")
			continue
		}
		tags = append(tags, C.GoString(C.VSL_tags[i]))
	}
	return tags
}

var (
	// Tags is an array of strings indexed by the Tag field of a
	// Record, classifying the Record's contents, as in the second
	// column of default varnishlog output. Examples are "ReqURL"
	// for a client request URL and "BerespStatus" for the HTTP
	// status of a backend response.
	//
	// If Tags[i] == "", then no tag exists for index i.
	Tags       = tags()
	nonprint   = regexp.MustCompile("[^[:print:]]")
	log2txchan = make(map[int]*(chan<- []*Transaction))
)

// A Record in the Varnish log, corresponding to a line of output from
// the varnishlog utility. Instances of this type are contained in the
// Transaction objects returned by Read() operations, and form the
// bulk of log data.
//
// The contents and format of log records are described in vsl(7):
// https://varnish-cache.org/docs/trunk/reference/vsl.html
type Record struct {
	// Indicates whether this record is part of a client or
	// backend log, or neither.
	Type RecordType

	// An index into the Tags array, classifying the contents of
	// this record, as in the second column of default varnishlog
	// output. For a Tag t, Tags[t] may be a string such as
	// "ReqURL" for a client request URL, "BerespStatus" for the
	// HTTP status of a backend response, and so forth.
	Tag uint8

	// The unique ID generated by Varnish for this log
	// transaction.
	VXID uint

	// The log payload is the specific data for this record. The
	// payload appears after the second column of default
	// varnishlog output.
	//
	// Unprintable bytes in the payload are rendered with the "%q"
	// verb for package fmt: "\xXX", where XX is the hex value of
	// the byte.
	Payload string
}

// A Transaction is a collection of related log records, typically as
// the log of a client or backend request and response. Transactions
// may also include data unrelated to requests and responses, such as
// backend health checks.
//
// Transactions are read from the log in groups, according to the
// Grouping selected for the log client, corresponding to the grouping
// modes set by varnishlog -g. Transactions in a group may form a
// hierarchy, in which case they are related by the parent VXID
// property -- a transaction's parent VXID is the VXID of the
// transaction that initiated it. Parent VXID 0 indicates that a
// transaction has no parent transaction.
//
// The depth of nesting in a hierarchy is represented by a
// transaction's Level property. Levels start at level 1, except for
// raw transactions, which always have level 0.
type Transaction struct {
	// Whether this is the log of a client request/response,
	// backend request/response, session and so forth.
	Type TxType

	// The cause of the event logged by this transaction.
	Reason Reason

	// The level of this transaction in a group. Always 1 for VXID
	// grouping, and always 0 for raw grouping.
	Level uint

	// The unique ID generated by Varnish for this transaction.
	VXID uint32

	// The ID of the parent transaction for this transaction in a
	// group. 0 if there is no parent. Always 0 for VXID and raw
	// groupings.
	ParentVXID uint32

	// The Records that make up the bulk of this transaction's
	// data.
	Records []*Record
}

// A Log instance is a client for the native logging interface.
type Log struct {
	vsl    *C.struct_VSL_data
	cursor *C.struct_VSL_cursor
	vslq   *C.struct_VSLQ
	grp    C.enum_VSL_grouping_e
	query  *C.char
}

// New returns a new and initialized instance of Log. Log instances
// should only be created through this method, since it performs
// initializations for the native library that are required for the
// other methods.
//
// New may return nil in rare cases, usually in an out of memory
// condition.
func New() *Log {
	log := new(Log)
	log.vsl = C.VSL_New()
	if log.vsl == nil {
		return nil
	}
	log.grp = C.VSL_g_vxid
	log.query = nil
	return log
}

// Release frees native resources associated with this client. You
// should always call Release when you're done using a Log client,
// otherwise there is risk of resource leakage. It is wise to make
// use of Go's defer mechanism:
//
//	vlog := log.New()
//	defer vlog.Release()
//	// ... do stuff with vlog
func (log *Log) Release() {
	C.VSL_Delete(log.vsl)
	if log.vslq != nil {
		C.VSLQ_Delete(&log.vslq)
	}
}

// Error returns the most recent error message from the native
// interface.
func (log *Log) Error() string {
	return C.GoString(C.VSL_Error(log.vsl))
}

// AttachFile attaches the client for reading from the binary log file
// located at path, as written by varnishlog -w.
//
// Returns nil if no file at that path is readable by the current
// process, if it is not a regular file, or if the native interface
// cannot successfully attach to the file.
func (log *Log) AttachFile(path string) error {
	info, err := os.Stat(path)
	if err != nil {
		return err
	}
	if !info.Mode().IsRegular() {
		return errors.New(path + " is not a regular file")
	}
	log.cursor = C.VSL_CursorFile(log.vsl, C.CString(path), 0)
	if log.cursor == nil {
		return log
	}
	return nil
}

func quote(s string) string {
	return fmt.Sprintf("%q", s)
}

//export publish
func publish(trans unsafe.Pointer, priv unsafe.Pointer) C.int {
	var gtxn []*Transaction
	status := C.int(0)
	txn := (*[1 << 30]*C.struct_VSL_transaction)(unsafe.Pointer(trans))
	n := int(*(*C.int)(priv))
	txChan, ok := log2txchan[n]
	if !ok {
		panic("unmapped transaction channel")
	}
	for i := 0; txn[i] != nil; i++ {
		vtx := txn[i]
		tx := new(Transaction)
		tx.Type = TxType(C.txtype(vtx))
		tx.VXID = uint32(vtx.vxid)
		tx.ParentVXID = uint32(vtx.vxid_parent)
		tx.Reason = Reason(vtx.reason)
		tx.Level = uint(vtx.level)
		for {
			status = C.int(C.VSL_Next(vtx.c))
			if status <= 0 {
				break
			}
			tag := C.tag(vtx.c)
			if tag < 0 || int(tag) >= len(Tags) {
				panic("tag index out of range")
			}
			if Tags[tag] == "" {
				panic("unknown tag")
			}
			rec := new(Record)
			rec.Tag = uint8(tag)
			rec.Payload = C.GoStringN(C.cdata(vtx.c), C.len(vtx.c))
			if C.unsafe(tag) != 0 || C.binary(tag) != 0 {
				nonprint.ReplaceAllStringFunc(rec.Payload,
					quote)
			}
			switch {
			case C.backend(vtx.c) != 0:
				rec.Type = Backend
			case C.client(vtx.c) != 0:
				rec.Type = Client
			default:
				rec.Type = None
			}
			rec.VXID = uint(C.vxid(vtx.c))
			tx.Records = append(tx.Records, rec)
		}
		gtxn = append(gtxn, tx)
	}
	*txChan <- gtxn
	return status
}

// Possible values for the Status field of a RWStatus object,
// indicating the error or end condition of a Varnish log read/write
// sequence.
const (
	// Write error.
	WriteErr = StatusType(C.vsl_e_write)

	// IO error while reading.
	IOErr = StatusType(C.vsl_e_io)

	// Overrun indicates that the write position of a live Varnish
	// instance, writing into the ring buffer, has cycled around
	// and overtaken the read position of the Log client.
	Overrun = StatusType(C.vsl_e_overrun)

	// The log was closed or abandoned (usually because the Varnish
	// child process stopped).
	Abandoned = StatusType(C.vsl_e_abandon)

	// The end of a binary log file was reached.
	EOF = StatusType(C.vsl_e_eof)

	// The end of a live Varnish log was reached (there is
	// currently no new log data).
	EOL = StatusType(C.vsl_end)
)

// StatusType is the type of the Status field of a RWStatus object,
// and classifies the error or end condition of a log read/write
// sequence. Possible values for StatusType are shown in the constants
// above.
type StatusType int8

// RWStatus indicates the error or end condition of a log read/write
// sequence. A pointers to an object of this type is sent on the
// status channel of a Read operation.
type RWStatus struct {
	// A classifier for the error or end condition.
	Status StatusType
}

// Error returns a string describing the RWStatus.
//
// Some of the strings returned, such as for EOF or EOL, do not
// necessarily describe an error.
func (rdstatus *RWStatus) Error() string {
	switch StatusType(rdstatus.Status) {
	case WriteErr:
		return "write error"
	case IOErr:
		return "I/O read error"
	case Overrun:
		return "Log overrun"
	case Abandoned:
		return "Log was abandoned or closed"
	case EOF:
		return "End of file"
	case EOL:
		return "End of log"
	default:
		panic("invalid read status")
	}
}

// Read starts a goroutine that reads from the currently attached log
// source, and sends the results as groups of pointers to Transaction
// on the txChan channel. When the read process encounters an end or
// error condition, a pointer to a RWStatus is sent on the errChan
// channel indicating the condition.
//
// Returns nil if either of txChan or statusChan is nil, or if the
// native library could not initialize the read.
func (log *Log) Read(txChan chan<- []*Transaction,
	statusChan chan<- *RWStatus) error {

	if txChan == nil {
		return errors.New("txChan is nil")
	}
	if statusChan == nil {
		return errors.New("statusChan is nil")
	}
	log.vslq = C.VSLQ_New(log.vsl, &log.cursor, log.grp, log.query)
	if log.vslq == nil {
		return log
	}
	n := len(log2txchan) + 1
	log2txchan[n] = &txChan
	go func(vslq *C.struct_VSLQ, n int) {
		var status C.int
		for status = C.vsl_more; status == C.vsl_more; {
			status = C.dispatch(vslq, C.int(n))
		}
		close(txChan)
		delete(log2txchan, n)
		rdstatus := new(RWStatus)
		rdstatus.Status = StatusType(status)
		statusChan <- rdstatus
		close(statusChan)
	}(log.vslq, n)
	return nil
}
